//---------------------------------------------------------------------------
// JCAMP-DX Leseprogramm
// (C) Copyright, Peter Lampen, ISAS, Dortmund, 07. Sept. 2004
//---------------------------------------------------------------------------

#pragma hdrstop

#include "jcamp1.h"



/*

//---------------------------------------------------------------------------

#pragma argsused
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fwdreason, LPVOID lpvReserved)
{ return 1;
}
//---------------------------------------------------------------------------

// Einen JCAMP-DX File lesen und die XY-Daten als ASCII-Tabelle speichern
int AsciiExport(const char *InputFile, const char *OutputFile,
  const char *separator)
{ int error;
  Jcamp jcamp;

  error = jcamp.ReadDxFile(InputFile);
  if (error) return error;

  error = jcamp.FindRecords();
  if (error) return error;

  error = jcamp.FindBlocks();
  if (error) return error;

  error = jcamp.ReadSpectra();
  if (error) return error;

  error = jcamp.AsciiOutput(OutputFile, separator);
  if (error) return error;

  return 0;
}
//---------------------------------------------------------------------------

// Aufruf der DLL in Origin: DLL jcamp dx2org %H <path\filename.dx>
int PASCAL DX2ORG(HWND hWnd, LPSTR lpString)
{ int error;
  char InputFile[256], WorkSheet[NAME_SIZE];
  Jcamp jcamp;

  // In lpString sind die Namen des Worksheets und des JCAMP-DX Files
  // gespeichert
  istrstream input(lpString);
  input >> WorkSheet >> ws;     // Leerzeichen entfernen
  input.getline(InputFile, sizeof(InputFile));

  error = jcamp.ReadDxFile(InputFile);
  if (error) return error;

  error = jcamp.FindRecords();
  if (error) return error;

  error = jcamp.FindBlocks();
  if (error) return error;

  error = jcamp.ReadSpectra();
  if (error) return error;

  error = jcamp.OriginExport(hWnd, WorkSheet);
  if (error) return error;

  return 0;
}


*/


//---------------------------------------------------------------------------

DxFile::~DxFile(void)                   // Destruktor
{ if (line != NULL)
  { for (unsigned int i = 0; i < lines; i++)
      if (line[i] != NULL) delete[] line[i];
    delete[] line;
  }
}
//---------------------------------------------------------------------------

int DxFile::ReadDXFile(const char *InputFile)  // Einlesen des JCAMP-DX-Files
{ char line2[128];
  ifstream input;                       // Inputstream

  input.open(InputFile);
  if (! input) return Jcamp::INPUT_FILE_ERR;   // Fehler beim ffnen

  memset(line2, 0, sizeof(line2));

  for (lines = 0; input.good(); lines++)       // Die Zeilen zhlen
    input.getline(line2, sizeof(line2));

// ios_base::iostate err= ios_base::goodbit;


//  input.clear(0);			// EOF-Bit zurcksetzen
  input.clear();			// EOF-Bit zurcksetzen
  input.seekg(0);			// Zeiger auf den Dateianfang setzen

  line = new char *[lines];             // Array aus char-Pointern
  if (line == NULL)
  { input.close();
    return Jcamp::MEMORY_ERR;           // kein freier Speicher
  }

  for (unsigned int i = 0; i < lines; i++)
  { input.getline(line2, sizeof(line2));

    line[i] = new char[strlen(line2) + 1];
    if (line[i] == NULL)
    { input.close();
      return Jcamp::MEMORY_ERR;
    }

    strcpy(line[i], line2);
  }

  input.close();			// den File schlieen
  return 0;
}
//---------------------------------------------------------------------------

// Die Zeilen aufbereiten und Kommentare lschen
void DxFile::MakeUp(void) const
{ char *ptr, *ptr2;

  for (unsigned int i = 0; i < lines; i++)
  { // Tabulatoren durch Leerzeichen ersetzen
    while ((ptr = strchr(line[i], '\t')) != NULL) *ptr = ' ';

    // fhrende Leerzeichen lschen
    for (ptr = line[i]; *ptr == ' '; ptr++);
    if (ptr != line[i]) strcpy(line[i], ptr);

    // $$-Kommentare am Zeilenende lschen
    if ((ptr = strstr(line[i], "$$")) != NULL)
      memset(ptr, 0, strlen(ptr));

    // Leerzeichen am Zeilenende lschen, aber nicht vor dem Zeilenanfang
    if (strlen(line[i]) > 0 && line[i][0] != '\0')
    { for (ptr = line[i] + strlen(line[i]) - 1; ptr >= line[i] && *ptr == ' ';
        ptr--) *ptr = '\0';
    }

    if (strncmp(line[i], "##", 2) == 0)         // Zeile mit ##Label=
    { ptr = line[i] + 2;			// "##" berspringen
      while (*ptr != '=' && *ptr != '\0')
      { *ptr = toupper(*ptr);			// Umwandeln in Grobuchstaben

	if (strchr(" -/_", *ptr) != NULL)
          strcpy(ptr, ptr+1);                   // Whitspace lschen
	else
          ptr++;
      }

      if (*ptr == '=')		// Leerzeichen nach dem '=' lschen
      { ptr++;
        for (ptr2 = ptr; *ptr2 == ' '; ptr2++);
        if (ptr2 != ptr) strcpy(ptr, ptr2);
      }
    }
  }
  return;
}
//---------------------------------------------------------------------------

// Initialisierung der Klassenvariablen Record::dxfile
DxFile *Record::dxfile = NULL;

int Record::SetRecord(const unsigned int first, const unsigned int li,
  const char *lab)
{ FirstLine = first;
  lines = li;

  label = new char[strlen(lab) + 1];
  if (label == NULL) return Jcamp::MEMORY_ERR;
  else strcpy(label, lab);
  return 0;
}
//---------------------------------------------------------------------------


Spectrum::Spectrum(void):               // Konstruktor
  // 3.1 Core Fixed Header Information
  Title(NULL), Version(NaN), DataType(0), DataClass(0), Origin(NULL),
  Owner(NULL),

  // 3.2 Core Variable Header Information
  BlockId(0), IMSPressure(NaN), CarrierGas(NULL), DriftGas(NULL),
  IonPolarity(NULL), IonizationMode(NULL), ShutterOpeningTime(NaN),

  // 3.3 Core Data und 3.4 Core Data Table
  XUnits(NULL), YUnits(NULL), FirstX(NaN), LastX(NaN), FirstY(NaN),
  XFactor(NaN), YFactor(NaN), NPoints(0), y(NULL),

  // 4.1 Global Notes
  year(0), month(0), day(0), hour(0), minute(0), second(NaN), timezone(0),
  SourceReference(NULL), CrossReference(NULL), SampleDescription(NULL),
  CASName(NULL), IUPACName(NULL), Names(NULL), Molform(NULL),
  CASRegistryNo(NULL), Concentrations(NULL), SpectrometerDataSystem(NULL),
  InstParameters(NULL), DataProcessing(NULL), Comment(NULL), XLabel(NULL),
  YLabel(NULL), MaxX(NaN), MinX(NaN), MaxY(NaN), MinY(NaN),

  // 4.2 Data-type Specific Notes
  ReducedMobility(NULL), IonizationEnergy(NaN), DriftGasFlow(NaN),
  CarrierGasMoisture(NaN), IonizationSource(NULL), ShutterGridPotential(NaN),
  RepetitionRate(NaN), Averages(0),

  error(0)

  { // 3.2 Core Variable Header Information
    ElectricField[0]  = NaN;
    ElectricField[1]  = NaN;
    IMSTemperature[0] = NaN;
    IMSTemperature[1] = NaN;

  // 4.2 Data-type Specific Notes
    memset(IonizationChamberGeometry, 0, sizeof(IonizationChamberGeometry));
    IonizationChamber[0] = NaN;
    IonizationChamber[1] = NaN;
    IonizationChamber[2] = NaN;
    memset(DriftChamberGeometry, 0, sizeof(DriftChamberGeometry));
    DriftChamber[0] = NaN;
    DriftChamber[1] = NaN;
    DriftChamber[2] = NaN;
    CarrierGasFlow[0] = NaN;
    CarrierGasFlow[1] = NaN;
  }
//---------------------------------------------------------------------------

Spectrum::~Spectrum(void)		// Destruktor
{ if (y != NULL) delete[] y;
}
//---------------------------------------------------------------------------



// Die Records auswerten
int Spectrum::ReadRecord(const Record *record)
{ int error2 = 0;
  char *line = record->GetLine(0), *label = record->GetLabel();

  // 3.1 Core Fixed Header Information
  if (strcmp(label, "TITLE") == 0)
    Title = record;
  else if (strcmp(label, "JCAMPDX") == 0)
    Version = GetFloat(line);
  else if (strcmp(label, "DATATYPE") == 0)
    DataType = ReadDataType(line);
  else if (strcmp(label, "DATACLASS") == 0)
    ReadDataClass(line);
  else if (strcmp(label, "ORIGIN") == 0)
    Origin = record;
  else if (strcmp(label, "OWNER") == 0)
    Owner = record;

  // 3.2 Core Variable Header Information
  else if (strcmp(label, "BLOCKID") == 0)
    BlockId = GetUnsignedInt(line);
  else if (strcmp(label, ".IMSPRESSURE") == 0)
    IMSPressure = GetFloat(line);
  else if (strcmp(label, ".CARRIERGAS") == 0)
    CarrierGas = record;
  else if (strcmp(label, ".DRIFTGAS") == 0)
    DriftGas = record;
  else if (strcmp(label, ".ELECTRICFIELD") == 0)
    Get2Float(line, ElectricField);
  else if (strcmp(label, ".IONPOLARITY") == 0)
    IonPolarity = record;
  else if (strcmp(label, ".IONIZATIONMODE") == 0)
    IonizationMode = record;
  else if (strcmp(label, ".IMSTEMPERATURE") == 0)
    Get2Float(line, IMSTemperature);
  else if (strcmp(label, ".SHUTTEROPENINGTIME") == 0)
    ShutterOpeningTime = GetFloat(line);

  // 3.3 Core Data
  else if (strcmp(label, "XUNITS") == 0)
    XUnits = record;
  else if (strcmp(label, "YUNITS") == 0)
    YUnits = record;
  else if (strcmp(label, "FIRSTX") == 0)
    FirstX = GetDouble(line);
  else if (strcmp(label, "LASTX") == 0)
    LastX = GetDouble(line);
  else if (strcmp(label, "FIRSTY") == 0)
    FirstY = GetFloat(line);
  else if (strcmp(label, "XFACTOR") == 0)
    XFactor = GetDouble(line);
  else if (strcmp(label, "YFACTOR") == 0)
    YFactor = GetDouble(line);
  else if (strcmp(label, "NPOINTS") == 0)
    NPoints = GetUnsignedInt(line);

  // 3.4 Core Data Table
  else if (strcmp(label, "XYDATA") == 0)
    error2 = ReadXYData(record);

  // 4.1 Global Notes
  else if (strcmp(label, "DATE") == 0)
    ReadDate(line);
  else if (strcmp(label, "TIME") == 0)
    ReadTime(line);
  else if (strcmp(label, "LONGDATE") == 0)
    ReadLongDate(line);
  else if (strcmp(label, "SOURCEREFERENCE") == 0)
    SourceReference = record;
  else if (strcmp(label, "CROSSREFERENCE") == 0)
    CrossReference = record;
  else if (strcmp(label, "SAMPLEDESCRIPTION") == 0)
    SampleDescription = record;
  else if (strcmp(label, "CASNAME") == 0)
    CASName = record;
  else if (strcmp(label, "IUPACNAME") == 0)
    IUPACName = record;
  else if (strcmp(label, "NAMES") == 0)
    Names = record;
  else if (strcmp(label, "MOLFORM") == 0)
    Molform = record;
  else if (strcmp(label, "CASREGISTRYNO") == 0)
    CASRegistryNo = record;
  else if (strcmp(label, "CONCENTRATIONS") == 0)
    Concentrations = record;
  else if (strcmp(label, "SPECTROMETERDATASYSTEM") == 0)
    SpectrometerDataSystem = record;
  else if (strcmp(label, "INSTRUMENTALPARAMETERS") == 0)
    InstParameters = record;
  else if (strcmp(label, "DATAPROCESSING") == 0)
    DataProcessing = record;
  else if (strcmp(label, "COMMENT") == 0)
    Comment = record;
  else if (strcmp(label, "XLABEL") == 0)
    XLabel = record;
  else if (strcmp(label, "YLABEL") == 0)
    YLabel = record;
  else if (strcmp(label, "MAXX") == 0)
    MaxX = GetFloat(line);
  else if (strcmp(label, "MINX") == 0)
    MinX = GetFloat(line);
  else if (strcmp(label, "MAXY") == 0)
    MaxY = GetFloat(line);
  else if (strcmp(label, "MINY") == 0)
    MinY = GetFloat(line);

  // 4.2 Data-type Specific Notes
  else if (strcmp(label, ".REDUCEDMOBILITY") == 0)
    ReducedMobility = record;
  else if (strcmp(label, ".IONIZATIONENERGY") == 0)
    IonizationEnergy = GetFloat(line);
  else if (strcmp(label, ".IONIZATIONCHAMBER") == 0)
    GetStringFloat(line, IonizationChamberGeometry, IonizationChamber);
  else if (strcmp(label, ".DRIFTCHAMBER") == 0)
    GetStringFloat(line, DriftChamberGeometry, DriftChamber);
  else if (strcmp(label, ".CARRIERGASFLOW") == 0)
    Get2Float(line, CarrierGasFlow);
  else if (strcmp(label, ".DRIFTGASFLOW") == 0)
    DriftGasFlow = GetFloat(line);
  else if (strcmp(label, ".CARRIERGASMOISTURE") == 0)
    CarrierGasMoisture = GetFloat(line);
  else if (strcmp(label, ".IONIZATIONSOURCE") == 0)
    IonizationSource = record;
  else if (strcmp(label, ".SHUTTERGRIDPOTENTIAL") == 0)
    ShutterGridPotential = GetFloat(line);
  else if (strcmp(label, ".REPETITIONRATE") == 0)
    RepetitionRate = GetFloat(line);
  else if (strcmp(label, ".AVERAGES") == 0)
    Averages = GetInt(line);

  return error2;
}
//---------------------------------------------------------------------------

const char *Spectrum::GetTimeString(void)   // Uhrzeit
{ memset(TimeString, 0, sizeof(TimeString));

  ostrstream output(TimeString, sizeof(TimeString));
  output << setfill('0') << setw(2) << hour << ":" << setw(2)
    << minute << ":" << setw(2) << int(second);

  return TimeString;
}
//---------------------------------------------------------------------------

// Statische Memberfunktion: Den Record DataType lesen
unsigned int Spectrum::ReadDataType(char *line)
{ // umwandeln in Grobuchstaben
  for (char *ptr = line; *ptr != '\0'; ptr++) *ptr = toupper(*ptr);

  if (strcmp(line, "LINK") == 0)
    return LINK;
  else if (strcmp(line, "INFRARED SPECTRUM") == 0)
    return INFRARED_SPECTRUM;
  else if (strcmp(line, "INFRARED PEAK TABLE") == 0)
    return INFRARED_PEAK_TABLE;
  else if (strcmp(line, "INFRARED INTERFEROGRAM") == 0)
    return INFRARED_INTERFEROGRAM;
  else if (strcmp(line, "INFRARED TRANSFORMED SPECTRUM") == 0)
    return INFRARED_TRANSFORMED_SPECTRUM;
  else if (strcmp(line, "RAMAN SPECTRUM") == 0)
    return RAMAN_SPECTRUM;
  else if (strcmp(line, "NMR FID") == 0)
    return NMR_FID;
  else if (strcmp(line, "NMR SPECTRUM") == 0)
    return NMR_SPECTRUM;
  else if (strcmp(line, "NMR PEAK TABLE") == 0)
    return NMR_PEAK_TABLE;
  else if (strcmp(line, "NMR PEAK ASSIGNMENTS") == 0)
    return NMR_PEAK_ASSIGNMENTS;
  else if (strcmp(line, "MASS SPECTRUM") == 0)
    return MASS_SPECTRUM;
  else if (strcmp(line, "CONTINUOUS MASS SPECTRUM") == 0)
    return CONTINUOUS_MASS_SPECTRUM;
  else if (strcmp(line, "ION MOBILITY SPECTRUM") == 0)
    return ION_MOBILITY_SPECTRUM;
  else if (strcmp(line, "IMS PEAK TABLE") == 0)
    return IMS_PEAK_TABLE;
  else if (strcmp(line, "IMS PEAK ASSIGNMENTS") == 0)
    return IMS_PEAK_ASSIGNMENTS;
  else return UNKNOWN;
}
//---------------------------------------------------------------------------

// Den Record ##DATA CLASS= lesen
void Spectrum::ReadDataClass(char *line)
{ // umwandeln in Grobuchstaben
  for (char *ptr = line; *ptr != '\0'; ptr++) *ptr = toupper(*ptr);

  if (strcmp(line, "XYDATA") == 0)
    DataClass = XYDATA;
  else if (strcmp(line, "XYPOINTS") == 0)
    DataClass = XYPOINTS;
  else if (strcmp(line, "PEAK TABLE") == 0)
    DataClass = PEAK_TABLE;
  else if (strcmp(line, "ASSIGNMENTS") == 0)
    DataClass = ASSIGNMENTS;
  else if (strcmp(line, "RADATA") == 0)
    DataClass = RADATA;
  else if (strcmp(line, "NTUPLES") == 0)
    DataClass = NTUPLES;
  else DataClass = 0;
}
//---------------------------------------------------------------------------

// Den Record ##DATE= lesen
void Spectrum::ReadDate(const char *line)
{ char slash;

  if (line != NULL && line[0] != '\0')
  { istrstream input(line);
    input >> year >> slash >> month >> slash >> day;
  }
  else
  { year = 0;
    month = 0;
    day = 0;
  }
  return;
}
//---------------------------------------------------------------------------

// Den Record ##TIME= lesen
void Spectrum::ReadTime(const char *line)
{ char slash, colon;

  if (line != NULL && line[0] != '\0')
  { istrstream input(line);
    input >> hour >> colon >> minute >> colon >> second;
  }
  else
  { hour = 0;
    minute = 0;
    second = 0.0;
  }
  return;
}
//---------------------------------------------------------------------------

// Den Record ##LONGDATE= lesen
void Spectrum::ReadLongDate(const char *line)
{ char slash, colon;

  if (line != NULL && line[0] != '\0')
  { istrstream input(line);
    input >> year >> slash >> month >> slash >> day >> hour >> colon >>
    minute >> colon >> second >> timezone;
  }
  else
  { year = 0;
    month = 0;
    day = 0;
    hour = 0;
    minute = 0;
    second = 0.0;
    timezone = 0;
  }
  return;
}
//---------------------------------------------------------------------------

// Den Record ##XYDATA= zeilenweise lesen
int Spectrum::ReadXYData(const Record *record)
{ double x, x2, DeltaX;
  int *IntY, *NewY, OldY;
  unsigned int i, NewPoints, lines, type, points;
  char *line;
  bool YValueCheck = false;

  if (NPoints > 2)
    DeltaX = (LastX - FirstX) / double(NPoints - 1);
  else
    DeltaX = 0.0;

  y = new float[NPoints];		// reelle Y-Werte
  IntY = new int[NPoints + 100];	// skalierte Y-Werte incl. Reserve
  if (y == NULL || IntY == NULL) return Jcamp::MEMORY_ERR;

  lines = record->GetLines();
  points = 0;
  for (i = 1; i < lines; i++)
  { NewY = IntY + points;
    line = record->GetLine(i);
    NewPoints = ReadXYLine(line, &x, NewY, &type);

    if (YValueCheck && (NewY[0] != OldY))
      error |= Y_VALUE_CHK_ERR;

    if (type == DIF)   		        // den letzten Punkt doppelt lesen
    { OldY = NewY[NewPoints - 1];
      NewPoints--;
      YValueCheck = true;
    }
    else
      YValueCheck = false;

    x2 = (FirstX + DeltaX * double(points)) / XFactor;
    if (fabs(x - x2) > fabs(DeltaX / 2.0))
      error |= X_SEQU_CHK_ERR;          // X-Sequence-Check Fehler
    points += NewPoints;
    if (points > NPoints) break;        // Error: zu viele Y-Werte
    // Elegant ist das nicht ..., aber besser als nichts.
  }

  for (i = 0; i < NPoints; i++)
    y[i] = float(IntY[i]) * YFactor;    // die reelen Y-Werte berechnen
  if (IntY != NULL) delete[] IntY;

  if (fabs(y[0] - FirstY) > YFactor) error |= FIRSTY_ERR;
  if (points != NPoints) error |= NPOINTS_ERR;

  return 0;
}
//---------------------------------------------------------------------------

// Eine Zeile des Records ##XYDATA= lesen
unsigned int Spectrum::ReadXYLine(char *line, double *x, int *y,
  unsigned int *type)
{ int dif;
  unsigned int i = 0, i0, type2, dup;
  char string[82], *ptr1 = line, *ptr2 = string;

  // den X-Wert bestimmen und ungltige Zeichen berspringen
  memset(string, 0, sizeof(string));
  while ((type2 = AsdfType(*ptr1)) == ILLEGAL) ptr1++;
  if (type2 == EOL) return 0;	        // leere Zeile
  *ptr2 = *ptr1;			// das erste Zeichen kopieren
  for (ptr1++, ptr2++; (isdigit(*ptr1) || (*ptr1 == '.')); ptr1++, ptr2++)
  { *ptr2 = *ptr1;
  }
  *x = GetDouble(string);

  while (*ptr1 != '\0')			// die Y-Werte lesen
  { memset(string, 0, sizeof(string));
    ptr2 = string;
    while ((type2 = AsdfType(*ptr1)) == ILLEGAL) ptr1++;

    switch (type2)
    { case DIGIT:
	for (; isdigit(*ptr1); ptr1++, ptr2++) *ptr2 = *ptr1;
 	y[i] = GetInt(string);
	*type = DIGIT;
	break;

      case SIGN:
        *ptr2 = *ptr1;                  // Das Vorzeichen kopieren
	for (ptr1++, ptr2++; isdigit(*ptr1); ptr1++, ptr2++) *ptr2 = *ptr1;
	y[i] = GetInt(string);
	*type = DIGIT;
	break;

      case SQZ:
        *ptr2 = *ptr1;                  // Die Pseudo-Ziffer kopieren
	for (ptr1++, ptr2++; isdigit(*ptr1); ptr1++, ptr2++) *ptr2 = *ptr1;
	y[i] = GetInt(Asdf2Digit(string));
	*type = SQZ;
	break;

      case DIF:
        *ptr2 = *ptr1;                  // Die Pseudo-Ziffer kopieren
	for (ptr1++, ptr2++; isdigit(*ptr1); ptr1++, ptr2++) *ptr2 = *ptr1;
	dif = GetInt(Asdf2Digit(string));

	if (i > 0)
          y[i] = y[i-1] + dif;
	else
          y[i] = 0;                     // Fehler !!

	*type = DIF;
	break;

      case DUP:
        *ptr2 = *ptr1;                  // Die Pseudo-Ziffer kopieren
	for (ptr1++, ptr2++; isdigit(*ptr1); ptr1++, ptr2++) *ptr2 = *ptr1;
	dup = GetInt(Asdf2Digit(string));

	if (i > 0)
        { for (i0 = i; i < (i0 + dup - 1); i++)
	  { if (*type == DIF)
              y[i] = y[i-1] + dif;      // n-mal dif
	    else
              y[i] = y[i0-1];           // n-mal y[i-1]
	  }
        }
        else		                // Fehler !!
          for (; i < dup; i++) y[i] = 0;

	i--;
	break;

      case MISSING:                     // fehlender Y-Wert
	y[i] = 0;
	break;

      case EOL:
	break;				// Zeilenende erreicht
    }
    i++;
  }
  return i;
}
//---------------------------------------------------------------------------

// Den Type der Pseudo-Ziffer bestimmen
unsigned int Spectrum::AsdfType(const char c)
{ if (c >= '0' && c <= '9') return DIGIT;
  else if (c == '@' || (c >= 'A' && c <= 'I') || (c >= 'a' && c <= 'i'))
    return SQZ;
  else if (c == '%' || (c >= 'J' && c <= 'R') || (c >= 'j' && c <= 'r'))
    return DIF;
  else if ((c >= 'S' && c <= 'Z') || c == 's') return DUP;
  else if (c == '+' || c == '-') return SIGN;
  else if (c == '?') return MISSING;
  else if (c == '\0') return EOL;
  else return ILLEGAL;
}
//---------------------------------------------------------------------------

// Dekodieren der Pseudo-Ziffern
char *Spectrum::Asdf2Digit(char *string)
{ char *pos, buffer[256],
    *pseudo = "0123456789@ABCDEFGHIabcdefghi%JKLMNOPQRjklmnopqrSTUVWXYZs+-?";
  char *start[] =
    { "0",  "1",  "2",  "3",  "4",  "5",  "6",  "7",  "8",  "9",  // digit
      "0",  "1",  "2",  "3",  "4",  "5",  "6",  "7",  "8",  "9",  // pos. SQZ
	   "-1", "-2", "-3", "-4", "-5", "-6", "-7", "-8", "-9",  // neg. SQZ
      "0",  "1",  "2",  "3",  "4",  "5",  "6",  "7",  "8",  "9",  // pos. DIF
	   "-1", "-2", "-3", "-4", "-5", "-6", "-7", "-8", "-9",  // neg. DIF
	    "1",  "2",  "3",  "4",  "5",  "6",  "7",  "8",  "9",  // DUP
	    "+",  "-",  "0"					  // +-?
    };

  memset(buffer, 0, sizeof(buffer));	// den Puffer lschen
  pos = strchr(pseudo, string[0]);	// Pseudo-Ziffer identifizieren
  if (pos != NULL) strcpy(buffer, start[pos-pseudo]);	// Pseudo-Ziffer ers.
  strcat(buffer, string+1);		// den Rest des Strings kopieren
  strcpy(string, buffer);		// den korr. String zurckkopieren
  return string;
}
//---------------------------------------------------------------------------

// Den Unsigned-Integer-Wert lesen
unsigned int Spectrum::GetUnsignedInt(const char *string) const
{ unsigned int value;

  if (string != NULL && string[0] != '\0')
  { istrstream input(string);
    input.setf(ios::dec, ios::basefield);           // Dezimalzahl
    input >> value;
  }
  else
    value = 0;
  return value;
}
//---------------------------------------------------------------------------

// Den Integer-Wert lesen
int Spectrum::GetInt(const char *string) const
{ int value;

  if (string != NULL && string[0] != '\0')
  { istrstream input(string);
    input.setf(ios::dec, ios::basefield);           // Dezimalzahl
    input >> value;
  }
  else
    value = 0;
  return value;
}
//---------------------------------------------------------------------------

// Den Double-Wert lesen
double Spectrum::GetDouble(const char *string) const
{ double value;

  if (string != NULL && string[0] != '\0')
  { istrstream input(string);
    input >> value;
  }
  else
    value = NaN;
  return value;
}
//---------------------------------------------------------------------------

// Den Float-Wert lesen
float Spectrum::GetFloat(const char *string) const
{ float value;

  if (string != NULL && string[0] != '\0')
  { istrstream input(string);
    input >> value;
  }
  else
    value = NaN;
  return value;
}
//---------------------------------------------------------------------------

// Zwei Float-Werte lesen
void Spectrum::Get2Float(const char *string, float *array) const
{ char komma;

  if (string != NULL && string[0] != '\0')
  { istrstream input(string);
    input >> array[0] >> komma >> array[1];
  }
  else
  { array[0] = NaN;
    array[1] = NaN;
  }
}
//---------------------------------------------------------------------------

void Spectrum::GetStringFloat(const char *string1, char *string2,
  float *array) const
{ char komma;

  if (string1 != NULL && string1[0] != '\0')
  { istrstream input(string1);
    input.get(string2, sizeof(string2), ',');        // bis zum Komma lesen
    for (unsigned int i = 0; i < 3; i++)
    { input >> komma >> array[i];
      if (input.fail())
      { array[i] = NaN;
        input.clear();
      }
    }
  }
  else
  { memset(string2, 0, sizeof(string2));
    array[0] = NaN;
    array[1] = NaN;
    array[2] = NaN;
  }
}



//---------------------------------------------------------------------------

Jcamp::~Jcamp(void)			// Destruktor
{ 

  if (spectrum != NULL) delete[] spectrum;
  if (block != NULL) delete[] block;
  if (record != NULL) delete[] record;
  if (dxfile != NULL) delete dxfile;

}
//---------------------------------------------------------------------------


int Jcamp::ReadDxFile(const char *InputFile)
{ int error;
  char *line;

  if (dxfile != NULL) delete dxfile;
  dxfile = new DxFile;
  if (dxfile == NULL) return MEMORY_ERR;

  error = dxfile->ReadDXFile(InputFile);        // Den File lesen
  if (error) return error;

  dxfile->MakeUp();			        // den Text aufbereiten

  line = dxfile->GetLine(0);                    // ##TITLE=...
  if (strncmp(line, "##TITLE=", 8) != 0)
    return FILE_FORMAT_ERR;                     // kein JCAMP-DX-File
  else
    return 0;
}
//---------------------------------------------------------------------------

int Jcamp::FindRecords(void)            // Die Labeled-Data-Records suchen
{ int error;
  unsigned int lines, len, l1, l2, ri = 0;
  char label[256], *line, *ptr1, *ptr2;

  if (dxfile == NULL) return OTHER_ERR;	// kein JCAMP-DX File vorhanden!

  if (record != NULL)			// alte Records lschen
  { delete[] record;
    records = 0;
    record = NULL;
  }

  lines = dxfile->GetLines();
  for (l1 = 0; l1 < lines; l1++)                // die Records zhlen
    if (strncmp(dxfile->GetLine(l1), "##", 2) == 0) records++;
  record = new Record[records];
  if (record == NULL) return MEMORY_ERR;        // Kein freier Speicher vorh.
  Record::SetDxFile(dxfile);                    // Die Klassenvariable setzen

  for (l1 = 0; l1 < lines; l1++)
  { line = dxfile->GetLine(l1);
    memset(label, 0, sizeof(label));

    if (strncmp(line, "##", 2) == 0)    // Labels suchen
    { ptr1 = line + 2;			// "##" berspringen
      ptr2 = strchr(line, '=');
      if (ptr2 == NULL) continue;	// illegales Label ohne '='
      len = ptr2 - ptr1;

      if (len == 0)                     // ##= ist ein Kommentar
        strcpy(label, "COMMENT");
      else
      { if (len >= sizeof(label)) len = sizeof(label) - 1;
        strncpy(label, ptr1, len);	// Das Label kopieren
      }

      // Das Label in der Textzeile lschen
      ptr2++;
      strcpy(line, ptr2);

      // Die Anzahl der Zeilen des Records bestimmen
      for (l2 = l1+1; l2 < lines && strncmp(dxfile->GetLine(l2), "##", 2) != 0;
        l2++);

      // Leerzeilen am Anfang und am Ende des Records ignorieren
      while (l2 > l1+1 && strcmp(dxfile->GetLine(l2-1), "") == 0) l2--;
      while (l1 < l2-1 && strcmp(dxfile->GetLine(l1), "") == 0) l1++;

      // Die Zeilen des Records speichern
      error = record[ri].SetRecord(l1, l2-l1, label);
      ri++;
      if (error) return error;
      l1 = l2 - 1;
    }
  }
  return 0;
}
//---------------------------------------------------------------------------

int Jcamp::FindBlocks(void)             // Die Blcke identifizieren
{ unsigned int ri;
  int bi, bi2;
  char *label;

  if (record == NULL) return OTHER_ERR;	// keine LDRs vorhanden

  if (block != NULL)			// alte Blcke lschen
  { delete[] block;
    blocks = 0;
    block = NULL;
  }

  for (ri = 0; ri < records; ri++)	// die Blcke zhlen
    if (strcmp(record[ri].GetLabel(), "TITLE") == 0) blocks++;
  block = new Block[blocks];
  if (block == NULL) return MEMORY_ERR;

  bi = -1;
  for (ri = 0; ri < records; ri++)
  { label = record[ri].GetLabel();

    if (strcmp(label, "TITLE") == 0)
      block[++bi].SetFirstRecord(record + ri);

    else if (strcmp(label, "END") == 0)
    { if (block[bi].GetLastRecord() == NULL) // Den akt. Daten-Block schlieen
      	block[bi].SetLastRecord(record + ri);
      else		// Den letzten Link-Block suchen und schlieen
      { for (bi2 = bi; bi2 >= 0 && block[bi2].GetLastRecord() != NULL; bi2--);
	if (bi2 >= 0) block[bi2].SetLastRecord(record + ri);
      }
    }

    else if ((strcmp(label, "DATATYPE") == 0) &&
      (block[bi].GetLastRecord() == NULL))
      block[bi].SetDataType(Spectrum::ReadDataType(record[ri].GetLine(0)));
  }

  //SH wenn Endmarke (##END=) fehlt, Flickwerk fr unsicheren Parser
  for (bi2 = bi; bi2 >= 0; bi2--)
  { if ( block[bi].GetLastRecord() == NULL )
      block[bi2].SetLastRecord(record + records-1);
  }
  return 0;
}
//---------------------------------------------------------------------------

int Jcamp::ReadSpectra(void)            // die Spektren lesen
{ int error;
  unsigned int bi, si;
  const Record *rec;

  if (block == NULL) return OTHER_ERR;	// keine Blcke vorhanden

  if (spectrum != NULL)			// alte Spektren lschen
  { delete[] spectrum;
    spectra = 0;
    spectrum = NULL;
  }

  for (bi = 0; bi < blocks; bi++)	// die Spektren zhlen
    if (block[bi].GetDataType() > Spectrum::LINK) spectra++;
  spectrum = new Spectrum[spectra];
  if (spectrum == NULL) return MEMORY_ERR;

  for (si = 0, bi = 0; bi < blocks; bi++)	// Schleife ber die Blcke
  { if (block[bi].GetDataType() <= Spectrum::LINK) continue;   // LINK-Block

    for (rec = block[bi].GetFirstRecord(); rec <= block[bi].GetLastRecord();
      rec++)
    { error = spectrum[si].ReadRecord(rec);
      if (error) return error;
    }
    si++;
  }
  return 0;
}
//---------------------------------------------------------------------------

// Die Daten im ASCII-Format ausgeben
int Jcamp::AsciiOutput(const char *OutputFile, const char *separator)
{ unsigned int i, si, npoints;
  ofstream output;                          // Output-Stream

  if (spectrum == NULL) return OTHER_ERR;   // keine Spektren vorhanden

  output.open(OutputFile);                  // den Ausgabefile ffnen
  if (! output) return OUTPUT_FILE_ERR;

  npoints = spectrum[0].GetNPoints();
  for (i = 0; i < npoints; i++)
  { output << spectrum[0].GetX(i);
    for (si = 0; si < spectra; si++)
      output << separator << spectrum[si].GetY(i);
    output << endl;
  }
  return 0;
}
//---------------------------------------------------------------------------

// Die Daten nach Origin exportieren
int Jcamp::OriginExport(HWND hWnd, const char *WorkSheet) const
{ double *data;
  char command[256];
  unsigned int i, j, npoints;
  WPARAM wParam;
  LRESULT lResult;

  if (spectrum == NULL) return OTHER_ERR;   // keine Spektren vorhanden

  // Das Label des Worksheets definieren
  sprintf(command, "%s!page.label$= %s", WorkSheet, spectrum[0].GetTitle());
  lResult = SendMessage(hWnd, WM_USER, 0, (LPARAM) command);
  if (lResult != 0) return ORIGIN_ERR;

  // Titel und Label des Worksheets anzeigen
  sprintf(command, "%s!page.title= 3", WorkSheet);
  lResult = SendMessage(hWnd, WM_USER, 0, (LPARAM) command);
  if (lResult != 0) return ORIGIN_ERR;

  // Falls notwendig, Spalten zum Worksheet hinzufgen
  if (spectra > 1)
  { sprintf(command, "worksheet -a %d", spectra - 1);
    lResult = SendMessage(hWnd, WM_USER, 0, (LPARAM) command);
    if (lResult != 0) return ORIGIN_ERR;
  }

  // Die Spalten in das double-Format bringen
  for (i = 1; i < spectra + 2; i++)
  { sprintf(command, "worksheet -f %d 1", i);
    lResult = SendMessage(hWnd, WM_USER, 0, (LPARAM) command);
    if (lResult != 0) return ORIGIN_ERR;
  }

  // Die Lnge der Spalten definieren
  npoints = spectrum[0].GetNPoints();
  sprintf(command, "set %s -e %d", WorkSheet, npoints);
  lResult = SendMessage(hWnd, WM_USER, 0, (LPARAM) command);
  if (lResult != 0) return ORIGIN_ERR;

  // Die Label der Spalten definieren
  for (i = 0; i < spectra; i++)
  { sprintf(command, "%s!wks.col%d.label$= %s", WorkSheet, i + 2,
      spectrum[i].GetTimeString());
    lResult = SendMessage(hWnd, WM_USER, 0, (LPARAM) command);
    if (lResult != 0) return ORIGIN_ERR;
  }

  // Das Worksheet ffnen
  lResult = SendMessage(hWnd, WM_DLL_OPEN_WKSHEET, 0, (LPARAM) WorkSheet);
  if (lResult != 0) return ORIGIN_ERR;

  // Die X-Spalte ffnen
  lResult = SendMessage(hWnd, WM_DLL_WKS_OPEN_COL, (WPARAM) 1, 0L);
  if (lResult == 0) return ORIGIN_ERR;
  else data = (double FAR *) lResult;

  // Die Daten exportieren
  for (i = 0; i < npoints; i++) data[i] = spectrum[0].GetX(i);

  // Die Y-Spalten ffnen, die Spektren exportieren und die Spalten schlieen
  for (i = 0; i < spectra; i++)
  { lResult = SendMessage(hWnd, WM_DLL_WKS_OPEN_COL, (WPARAM) (i + 2), 0L);
    if (lResult == 0) return ORIGIN_ERR;
    else data = (double FAR *) lResult;

    for (j = 0; j < npoints; j++) data[j] = spectrum[i].GetY(j);
  }

  // Das Worksheet schlieen
  SendMessage(hWnd, WM_DLL_CLOSE_WKSHEET, 1, 0L);
  return 0;
}
//---------------------------------------------------------------------------

