//---------------------------------------------------------------------------
// JCAMP-DX Leseprogramm
// (C) Copyright, Peter Lampen, ISAS, Dortmund, 07. Sept. 2004
//---------------------------------------------------------------------------

#ifndef jcamp1H
#define jcamp1H

#include <windows.h>

using namespace std;
#include <iostream>
#include <fstream.h>
#include <fstream>

#include <strstream>  // for istrstream
#include <iomanip.h>
#include <math.h>

#include <limits.h>         // for USHRT_MAX
#include <stdlib.h>         // for calloc, min
#include <string.h>         // for memset, memcpy
#include <fstream>        // for ifstream

#include "ORGDLL.H"       // Origin Interface

const float NaN = 3.4e+38f;    // not-a-number

//---------------------------------------------------------------------------

class DxFile	// ASCII-Text, bestehend aus Zeilen variabler Lnge
{ 
      public:
    DxFile(void): lines(0), line(NULL) { }	// Konstruktor
    ~DxFile(void);				// Destruktor
    int ReadDXFile(const char *InputFile);      // den Text einlesen
    void MakeUp(void) const;		        // die Text-Zeilen aufbereiten
    unsigned int GetLines(void) const { return lines; }
    char *GetLine(const unsigned int i = 0) const
    { return (i < lines) ? line[i] : NULL; }

  private:
    unsigned int lines;			        // Anzahl der Textzeilen
    char **line;			        // 2D-Textarray
};


//---------------------------------------------------------------------------

class Record	// Labeled-Data-Record (LDR), bestehend aus Zeilen
{ public:
    Record(void): FirstLine(0), lines(0), label(NULL) { }
    ~Record(void) { if (label != NULL) delete[] label; } // Destruktor
    int SetRecord(const unsigned int first, const unsigned int li,
      const char *lab);
    unsigned int GetLines(void) const { return lines; }
    char *GetLine(const unsigned int i = 0) const
    { return (i < lines) ? dxfile->GetLine(FirstLine + i) : NULL; }
    char *GetLabel(void) const { return label; }
    static void SetDxFile(DxFile *const dx) { dxfile = dx; }

  private:
    unsigned int FirstLine, lines;
    char *label;
    static DxFile *dxfile;      // Klassenvariable
};
//---------------------------------------------------------------------------

class Block	// Link- oder Data-Block, bestehend aus LDRs
{ public:
    Block(void): DataType(0), FirstRecord(NULL), LastRecord(NULL) { }
    void SetFirstRecord(const Record *first) { FirstRecord = first; }
    void SetLastRecord(const Record *last) { LastRecord = last; }
    void SetDataType(const unsigned int type) { DataType = type; }
    const Record *GetFirstRecord(void) const { return FirstRecord; }
    const Record *GetLastRecord(void) const { return LastRecord; }
    int GetDataType(void) const { return DataType; }

  private:
    unsigned int DataType;
    const Record *FirstRecord, *LastRecord;
};
//---------------------------------------------------------------------------

class Spectrum	// spektrale Daten
{ public:
    // Definierte Werte fr DataType (Ergebnis der Fkt. ReadDataType())
    enum {UNKNOWN, LINK, INFRARED_SPECTRUM, INFRARED_PEAK_TABLE,
      INFRARED_INTERFEROGRAM, INFRARED_TRANSFORMED_SPECTRUM, RAMAN_SPECTRUM,
      NMR_FID, NMR_SPECTRUM, NMR_PEAK_TABLE, NMR_PEAK_ASSIGNMENTS,
      MASS_SPECTRUM, CONTINUOUS_MASS_SPECTRUM, ION_MOBILITY_SPECTRUM,
      IMS_PEAK_TABLE, IMS_PEAK_ASSIGNMENTS};

    Spectrum(void);				// Konstruktor
    ~Spectrum(void);				// Destruktor
    int ReadRecord(const Record *record);	// Die Daten eines LDRs lesen
    const char *GetTitle(void) const            // Titel des Spektrums
    { return (Title != NULL) ? Title->GetLine(0) : NULL; }
    unsigned int GetNPoints(void) const         // Anzahl der Punkte
    { return NPoints; }
    double GetX(const unsigned int i) const     // x[i] zurckgeben
    { return (FirstX + double(i) * (LastX - FirstX) / double(NPoints - 1)); }
    float GetY(const unsigned int i) const      // y[i] zurckgeben
    { return (i < NPoints && y != NULL) ? y[i] : 0.0; }
    const char *GetTimeString(void);            // Uhrzeit
    static unsigned int ReadDataType(char *line);

  private:
    // Definierte Werte fr die Variable DataClass
    enum {XYDATA=1, XYPOINTS, PEAK_TABLE, ASSIGNMENTS, RADATA, NTUPLES};

    // Pseudoziffern fr das ASDF-Format (Ergebnis der Fkt. AsdfType(c))
    enum {ILLEGAL, DIGIT, SIGN, SQZ, DIF, DUP, MISSING, EOL};

    // Fehlermeldungen
    enum {X_SEQU_CHK_ERR=1, Y_VALUE_CHK_ERR=2, FIRSTY_ERR=4, NPOINTS_ERR=8};

    void ReadDataClass(char *line);
    void ReadDate(const char *line);
    void ReadTime(const char *line);
    void ReadLongDate(const char *line);
    int ReadXYData(const Record *record);	// Den Record ##XYDATA= lesen
    unsigned int ReadXYLine(char *line, double *x, int *y, unsigned int *type);
    unsigned int AsdfType(const char c);
    char *Asdf2Digit(char *string);
    unsigned int GetUnsignedInt(const char *string) const;
    int GetInt(const char *string) const;
    double GetDouble(const char *string) const;
    float GetFloat(const char *string) const;
    void Get2Float(const char *string, float *array) const;
    void GetStringFloat(const char *string1, char *string2, float *array)
      const;

    // 3.1 Core Fixed Header Information
    float Version;
    unsigned int DataType, DataClass;
    const Record *Title, *Origin, *Owner;

    // 3.2 Core Variable Header Information
    float IMSPressure, ElectricField[2], IMSTemperature[2], ShutterOpeningTime;
    unsigned int BlockId;
    const Record *CarrierGas, *DriftGas, *IonPolarity, *IonizationMode;

    // 3.3 Core Data und 3.4 Core Data Table
    double FirstX, LastX, XFactor, YFactor;
    float FirstY, *y;
    unsigned int NPoints;
    const Record *XUnits, *YUnits;

    // 4.1 Global Notes
    float second, MaxX, MinX, MaxY, MinY;
    int timezone;
    unsigned int year, month, day, hour, minute;
    char TimeString[20];
    const Record *SourceReference, *CrossReference, *SampleDescription,
      *CASName, *IUPACName, *Names, *Molform, *CASRegistryNo, *Concentrations,
      *SpectrometerDataSystem, *InstParameters, *DataProcessing, *Comment,
      *XLabel, *YLabel;

    // 4.2 Data-type Specific Notes
    float IonizationEnergy, IonizationChamber[3], DriftChamber[3],
      CarrierGasFlow[2], DriftGasFlow, CarrierGasMoisture,
      ShutterGridPotential, RepetitionRate;
    unsigned int Averages;
    char IonizationChamberGeometry[5], DriftChamberGeometry[5];
    const Record *ReducedMobility, *IonizationSource;

    int error;
};
//---------------------------------------------------------------------------

class Jcamp	// JCAMP-DX File
{ 
  public:
    // Fehlermeldungen
    enum {MEMORY_ERR=1, INPUT_FILE_ERR, OUTPUT_FILE_ERR, FILE_FORMAT_ERR, ORIGIN_ERR, OTHER_ERR};

    Jcamp(void): records(0), blocks(0), spectra(0), dxfile(NULL), record(NULL), block(NULL), spectrum(NULL) { }
    ~Jcamp(void);

    int ReadDxFile(const char *InputFile);      // den JCAMP-DX-File einlesen
    int FindRecords(void);		// die LDRs identifizieren
    int FindBlocks(void);		// die Blcke identifizieren
    int ReadSpectra(void);		// die spektralen Daten lesen
    int AsciiOutput(const char *OutputFile, const char *separator);
    int OriginExport(HWND hWnd, const char *WorkSheet) const; // n. Origin exp.

  private:
    unsigned int records, blocks, spectra;
    DxFile *dxfile;			// gesamter JCAMP-DX-File
    Record *record;			// Array aus Labeled-Data-Records
    Block *block;			// Array aus Link- oder Daten-Blcke
    Spectrum *spectrum;			// Array aus Spektren

};

//---------------------------------------------------------------------------


#endif

