#ifndef QUADSTAR_H
#define QUADSTAR_H

#include <QWidget>
#include <QtGui>

#include <qextserialport.h>

#include "ui_quadstar.h"

#include <SDL/SDL.h>
#undef main // permet d'viter des erreurs de conflits entre QT et SDL
#include <SDL/SDL_thread.h>
#include <SDL/SDL_mutex.h>

#include <windows.h> // pour le sleep

#include <jcamp1.h> // pour la conversion des fichiers

#include "plot.h" // pour l'affichage du graphique

#include "delegate.h" // pour le tableur

class QuadStar : public QMainWindow
{
    Q_OBJECT

public:
    QuadStar(QWidget *parent = 0);
    ~QuadStar();

    Ui::Quadstar ui;

  // page 1 : reception
  QextSerialPort *port; // sert a communiquer
  SDL_Thread *SDL_t1; // Thread SDL
  // les evenements servent pour l'affichage a l'ecran
  virtual bool event(QEvent *ev);
  QEvent *e1;
  int recup;
  int send_commande(QString); // envoi de commande
  int send_commande2R(QString); // affichage de ce qui est fait
  int send_commande2G(QString); // affichage de ce qui est fait
  int send_commande2B(QString); // affichage de ce qui est fait
  int send_commande3(QString); // affichage de ce qui est recu
  void reception(void); // reception des donnees
   QString messages_recus;
   QString message_a_afficher;
  QString name_file_JDX;

  // page 2 : conversion
  int affiche2B(QString la_commande); // affichage en bleu
  int affiche2R(QString la_commande); // affichage en rouge
  int affiche2V(QString la_commande); // affichage en vert
  void conversion(void); // convertir les fichiers
  void tableaudosage(void); // creation du tableau de dosage

  // page 3 : graphiques
  Plot visionGraph; // fenetre pour afficher les graphs des moteurs
  void enableZoomMode(bool);
  void chargement_valeurs(void); // chargement des valeurs
  void mode_zoom(void); // zoom / dezoom
  void name_graph(void); // donner un nom au graphique
  int recherche_pics(int min , int max, int num_courbe, double seuil); // recherche un pic dans l'interval min-max
  void recherche_pics_main(void); // recherche un pic dans l'interval min-max
  Zoomer *d_zoomer[2];
  QwtPlotPicker *d_picker ;
  QwtPlotPanner *d_panner ;

  // page 4 : tableur



public slots:

// page 1 : Reception
  void on_pushButton_2_clicked(); // Reception des donnees emises par le spectro
  void on_pushButton_3_clicked(); // sauvegarde des donnees dans un fichier

// page 3 : Graphiques
  void on_pushButton_5_clicked(); // affichage graph
  void on_pushButton_6_clicked(); // affichage 0%
  void on_pushButton_7_clicked(); // affichage 10%
  void on_pushButton_8_clicked(); // affichage % inconnu
  void on_pushButton_9_clicked(); // affichage 20%
  void on_pushButton_10_clicked(); // affichage A
  void on_pushButton_11_clicked(); // affichage B
  void on_pushButton_12_clicked(); // affichage C
  void on_pushButton_13_clicked(); // affichage D
  void on_pushButton_4_clicked(); // affichage A - CCL4
  void on_pushButton_14_clicked(); // affichage B - CCL4
  void on_pushButton_19_clicked(); // affichage C - CCL4
  void on_pushButton_20_clicked(); // affichage D - CCL4
  void on_pushButton_15_clicked(); // Impression
  void on_pushButton_16_clicked(); // Affichage background
  void on_pushButton_17_clicked(); // Affichage cuve vide
  void on_pushButton_18_clicked(); // Recadre
  void on_pushButton_clicked(); // Trouver les pics des solutions A, B, C et D

// variables
public:

// page 3 : Graphiques

  int state1;
  int state2;
  int state3;
  int state4;
  int state5;
  int state6;
  int state7;
  int state8;
  int state9;
  int state10;
  int state11;
  int state12;
  int state13;
  int state14;
  int statezoom;
  int state_pics;

// page 4 : Tableur
  QStandardItemModel *model;
  SpinBoxDelegate delegate;
  QModelIndex index;

};

#endif
