////////////////////////////////////////
// calcul equation courbe
long double CaptureImage::regPolynomiale(int degre,long double *coeffs,long double *coeffs_incertitude,long double *vals[2],int debut,int fin)
{
    assert(fin>debut); // le point de fin doit etre superieur au point de debut
    assert(degre>0); // degre doit etre superieur a 0

    if(!(degre<fin-debut)) // on verifie que l'on a suffisamment de points par rapport au degre
      return 0;  

    // fin-debut+1 = nombre de points de mesures pour les raies de reference. normalement, cela fait 16 points (9 + 7)

    enum{x,y}; // x=0 et y=1



    // Construction d'un tableau avec la somme des x, x^2
    long double *xPower=new long double[2*degre+1];
    for(int i=0;i<2*degre+1;i++)
    {
        xPower[i]=0;
    }    
    for(int j=debut;j<=fin;j++)
    {
        xPower[0]++;
        long double tmp=vals[x][j]; // ici, on a les valeurs en pixels de raies de reference
        // si 16 valeurs, alors cela va de 0 a 15


        for(int i=1;i<2*degre+1;i++)
        {
             // si de degre 3, alors cela va de 1 a 6
             xPower[i] = xPower[i] + tmp;
             tmp = tmp * vals[x][j];
        }
    }


    // Construction d'un tableau avec la somme de y, yx, yx2
    long double *yxPower=new long double[degre+1];
    for(int i=0;i<degre+1;i++)
    {
        yxPower[i]=0;
    }
    for(int j=debut;j<=fin;j++)
    {
        long double tmp=vals[y][j];
        for(int i=0;i<degre+1;i++)
        {
            yxPower[i]+=tmp;
            tmp*=vals[x][j];
        }
    }



    // Creation de la matrice X du systeme : X * A = Y
    double **matX = NULL;
    matX = create_matrice_carre(matX,degre+1);
//    matrice matX(degre+1);
    for(int ligne=0 ; ligne<degre+1 ; ligne++)
    {
        for(int colonne=0 ; colonne<degre+1 ; colonne++)
        {
//            matX(l,c)=xPower[degre*2-l-c];
              // format [colonne][ligne]
              matX[colonne][ligne]=xPower[degre*2-ligne-colonne];
        }
    }



    // Creation du vecteur Y du systeme : X * A = Y
    double **matY = NULL;
    matY = create_matrice(matY , 1 , degre+1);
    // format [colonne][ligne]
//    matrice matY(degre+1,1);
    for(int ligne=0 ; ligne<degre+1 ; ligne++)
    {
//        matY(l,0)=yxPower[degre-l];
          matY[0][ligne]=yxPower[degre-ligne];
    }


    // Creation du vecteur A du systeme : X * A = Y
    double **matX_inverse = NULL;
    matX_inverse = create_matrice_carre(matX_inverse,degre+1);
    inverse(matX , matX_inverse , degre+1 , degre+1);
    double **matA = NULL;
    matA = create_matrice(matA , 1 , degre+1);
    // on initialise la matrice A
	for (int ligne=0 ; ligne<degre+1 ; ligne++)
		for (int colonne=0 ; colonne<1 ; colonne++)
		  matA[colonne][ligne]=0;
    // on remplit la matrice
	for (int ligne=0 ; ligne<degre+1 ; ligne++)
	{
		for (int colonne=0 ; colonne<1 ; colonne++)
		{
			for (int k=0 ; k<degre+1 ; k++)
			{
				matA[colonne][ligne]=matX_inverse[k][ligne]*matY[colonne][k]+matA[colonne][ligne];
			}
		}
	}
//    matrice matA=matX.getInverse()*matY;


    // Affectation des coefficients du polynome
    for(int i=0;i<degre+1;i++)
    {
        coeffs[i]=matA[0][degre-i];  
    }    



    // Calcul du coeff de determination
    // Formule : r2=somme((newY-moyY)) / somme((oldY-moyY))
    long double R2=0;
    long double diviseur=0;
    long double moyY=yxPower[0]/xPower[0];
    long double valY;

    for(int i=debut;i<=fin;i++)
    {
        valY=0;
        for(int j=0;j<degre+1;j++)
        {
            valY+=coeffs[j]*pow(vals[x][i],j); 
        }
        R2+=pow((valY-moyY),2);
        diviseur+=pow((vals[y][i]-moyY),2);
    }

    if(diviseur!=0)
      R2/=diviseur;

    // Construction d'un tableau avec les incertitudes sur les coefficients du polynome
    for(int i=0;i<degre+1;i++)
    {
        coeffs_incertitude[i]= sqrt(matX_inverse[i][i] );  
    }    


    delete xPower;
    delete yxPower;
    

    // destruction des matrices
  // suppression de la matrice m2
  if (matX) 
  {
    for (int i=0; i<degre+1; ++i )
      delete matX[i];
    delete matX;
  }
  if (matX_inverse) 
  {
    for (int i=0; i<degre+1; ++i )
      delete matX_inverse[i];
    delete matX_inverse;
  }
  if (matY) 
  {
    for (int i=0; i<1; ++i )
      delete matY[i];
    delete matY;
  }
  if (matA) 
  {
    for (int i=0; i<1; ++i )
      delete matA[i];
    delete matA;
  }

    // Retourne le coefficient de determination
    return R2;

}


